﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace Text_Classification
{
    class Category
    {
        public delegate void WorkDone(int current, int max);
        public event WorkDone TokenizeDone;
        public event WorkDone StemmerDone;

        public Category(string folder)
        {
            this.folder = folder;
            int i = folder.LastIndexOf("\\");
            //category name will be folder name by default
            this.name = folder.Substring(i + 1);
        }

        public void Build()
        {
            extractVocabulary();
            //condprob will be calculated when we know 
            // the total num of terms in model
            // prior will be calculated when we know 
            // the total number of documents in model
        }

        public void DropTerm(object key)
        {
            this.vocabulary.Remove(key);
            this.condprob.Remove(key);
            this.tf_idf.Remove(key);
            this.documentFrequency.Remove(key);
        }

        public void CondProbCalc(int totalTerms)
        {
            // denom = Sum(Tct' + 1), 
            // t' element of Terms, 
            // Tc: concatnation of all documents in category
            // Tct': frequency of term t' in Tc
            // => denom = total number of terms + Sum(Tct')
            this.denom = totalTerms;
            // +sum(Tct')
            foreach (System.Collections.DictionaryEntry entry in this.vocabulary)
            {
                this.denom += (int)entry.Value;
            }

            foreach (System.Collections.DictionaryEntry entry in this.vocabulary)
            {
                this.condprob.Add(entry.Key, (1 + (int)entry.Value) / this.denom);
            }
        }

        // this void is for Model call when all documents in model collected
        public void PriorCalc(int countAllDocuments)
        {
            this.prior = (double)this.countDocuments / countAllDocuments;
        }

        private void extractVocabulary()
        {
            //get all file in folder
            string[] fileList = Directory.GetFiles(folder);
            //each file is a document
            //countDocuments in this category
            this.countDocuments = fileList.Length;
            Tokenizer tokenizer = new Tokenizer();
            foreach (string file in fileList)
            {
                TextReader tr = new StreamReader(file);
                string doc = tr.ReadToEnd();
                tr.Close();
                //Extract all vocabularies and their frequencies in this file 
                // and add to category vocabulary list
                // also count document frequency here
                vocabularyMerger(tokenizer.Partition(doc));
            }
        }

        // add vocabulary from a document to category and count document frequency for each term
        private void vocabularyMerger(System.Collections.SortedList words)
        {
            foreach (System.Collections.DictionaryEntry term in words)
            {
                if (this.vocabulary.ContainsKey(term.Key))
                {
                    // term exists in this.terms, increase frequency
                    this.vocabulary[term.Key] = (int)this.vocabulary[term.Key] + (int)term.Value;
                    this.documentFrequency[term.Key] = (int)this.documentFrequency[term.Key] + 1;
                }
                else
                {
                    // new term
                    this.vocabulary.Add(term.Key, term.Value);
                    this.documentFrequency.Add(term.Key, 1);
                }
            }
        }

        // Sum(Tct' + 1), t = 0 .. countDocuments - 1  
        public double Denom
        {
            get { return this.denom; }
        }

        public string Folder
        {
            get { return this.folder; }
        }

        public string Name
        {
            get { return this.name; }
        }

        public int CountDocuments
        {
            get { return this.countDocuments; }
        }

        public System.Collections.SortedList Vocabulary
        {
            get { return this.vocabulary; }
        }

        public System.Collections.SortedList CondProb
        {
            get { return this.condprob; }
        }

        public double Prior
        {
            get { return this.prior; }
        }

        public System.Collections.SortedList TF_IDF
        {
            get { return this.tf_idf; }
            set
            {
                this.tf_idf = value;
            }
        }

        public double TF_IDF_Threshold
        {
            get { return this.tf_idfThreshold; }
            set
            {
                this.tf_idfThreshold = value;
            }
        }

        public System.Collections.SortedList DocumentFreq
        {
            get { return this.documentFrequency; }
        }

        private int countDocuments = 0;
        private string folder;
        private string name;
        private double prior = 0;
        private double denom = 0;
        private System.Collections.SortedList vocabulary = new System.Collections.SortedList();
        private System.Collections.SortedList condprob = new System.Collections.SortedList();
        private System.Collections.SortedList tf_idf = new System.Collections.SortedList();
        private System.Collections.SortedList documentFrequency = new System.Collections.SortedList();
        private double tf_idfThreshold = int.MinValue;
    }
}
