﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.IO;
using System.Threading;

namespace Text_Classification
{
    class Model
    {
        public delegate void ShowProgress(int current, int max);
        public event ShowProgress ProgressIndicated;
        public event ShowProgress BigMovementIndicated;

        public Model(string folder)
        { 
            this.folder = folder;
            int i = folder.LastIndexOf("\\");
            //category name will be folder name by default
            this.name = folder.Substring(i + 1);
        }

        public void Build()
        {
            //Get all subfolder,
            string[] directories = Directory.GetDirectories(folder);
            this.countCategories = directories.Length;
            //each subfolder is one category
            int current = 0;
            foreach (string directory in directories)
            {
                Category cat = new Category(directory);
                //Generate tokens will take long time,
                //we need to show progress
                cat.TokenizeDone += this.progressIndicated;
                cat.Build();
                cat.TokenizeDone -= this.progressIndicated;
                //Finish building model for 1 category
                this.categories.Add(cat);
                this.vocabularyMerger(cat);
                this.countDocuments += cat.CountDocuments;
                current++;
                if (BigMovementIndicated != null)
                {
                    BigMovementIndicated(current, CountCategories);
                }
            }

            parameterCalculate();
        }

        private void parameterCalculate()
        {
            foreach (Category cat in categories)
            {
                cat.PriorCalc(this.CountDocuments);
                cat.CondProbCalc(this.terms.Count);
            }
            // calculate IDF for each term
            idfCalc();
            // calculate TF-IDF for each term
            tfIdfCalc();
        }

        private void tfIdfCalc()
        {
            foreach (Category cat in categories)
            {
                cat.TF_IDF = new System.Collections.SortedList();
                foreach (System.Collections.DictionaryEntry word in cat.Vocabulary)
                {
                    cat.TF_IDF.Add(word.Key, (double)cat.CondProb[word.Key] * (double)termsIDF[word.Key]);
                }
            }
        }

        private void dropTerms()
        {
            System.Collections.SortedList copy = 
                (System.Collections.SortedList)terms.Clone();
            foreach (System.Collections.DictionaryEntry entry in copy)
            {
                if ((int)entry.Value <= this.thresholdMin || (int)entry.Value >= this.thresholdMax) 
                {
                    terms.Remove(entry.Key);
                    foreach (Category cat in categories)
                    {
                        if (cat.Vocabulary.ContainsKey(entry.Key))
                            cat.DropTerm(entry.Key);
                    }
                }
            }
        }
        
        public int ThreshHoldMin
        {
            get { return this.thresholdMin;}
            set 
            {
                if (ThreshHoldMin < value)
                {
                    this.ThreshHoldMin = value;
                    dropTerms();
                }
            }
        }

        public int ThreshHoldMax
        {
            get { return this.thresholdMin;}
            set 
            {
                if (ThreshHoldMin < value)
                {
                    this.ThreshHoldMin = value;
                    dropTerms();
                }
            }
        }
        
        public bool SetThreshold(int min, int max)
        {
                bool threshHoldChanged = false;
                if (this.thresholdMin < min)
                {
                    this.thresholdMin = min;
                    threshHoldChanged = true;
                }
                if (this.thresholdMax > max)
                {
                    this.thresholdMax = max;
                    threshHoldChanged = true;
                }
                if (threshHoldChanged) 
                    dropTerms();
                return threshHoldChanged;
        }

        public void FinizeTermsNew(int threshold)
        {
            System.Collections.ArrayList tf_idfSortList = new System.Collections.ArrayList();
            double dropPoint;
            foreach (Category cat in categories)
                foreach (System.Collections.DictionaryEntry word in cat.Vocabulary)
                {
                    tf_idfSortList.Add((double)cat.TF_IDF[word.Key]);
                }
            tf_idfSortList.Sort();
            if (tf_idfSortList.Count > threshold)
            {
                dropPoint = (double)tf_idfSortList[tf_idfSortList.Count - 1 - threshold];
            }
            else
                dropPoint = (double)tf_idfSortList[0];


            System.Collections.SortedList copy =
                (System.Collections.SortedList)terms.Clone();
            int current = 0;
            foreach (System.Collections.DictionaryEntry entry in copy)
            {
                bool shouldDrop = true;
                // scan through categories if_tdf threshold
                foreach (Category cat in categories)
                {
                    if (cat.TF_IDF.ContainsKey(entry.Key))
                        if ((double)cat.TF_IDF[entry.Key] >= dropPoint)
                            shouldDrop = false;
                }
                if (shouldDrop)
                {
                    terms.Remove(entry.Key);
                    termsIDF.Remove(entry.Key);
                    DF.Remove(entry.Key);
                    foreach (Category cat in categories)
                    {
                        if (cat.Vocabulary.ContainsKey(entry.Key))
                            cat.DropTerm(entry.Key);
                    }
                }
                current++;
                if (BigMovementIndicated != null)
                {
                    BigMovementIndicated(current, Terms.Count);
                }

            }
        }

        public void FinizeTerms(int threshold)
        {
            foreach (Category cat in categories)
            {
                System.Collections.ArrayList tf_idfSortList = new System.Collections.ArrayList();
                foreach (System.Collections.DictionaryEntry word in cat.Vocabulary)
                {
                    tf_idfSortList.Add((double)cat.TF_IDF[word.Key]);
                }
                tf_idfSortList.Sort();
                if (tf_idfSortList.Count > threshold)
                {
                    cat.TF_IDF_Threshold = (double)tf_idfSortList[tf_idfSortList.Count - 1 - threshold];
                }
                else 
                    cat.TF_IDF_Threshold = (double)tf_idfSortList[0];
            }

            System.Collections.SortedList copy =
                (System.Collections.SortedList)terms.Clone();
            int current = 0;
            foreach (System.Collections.DictionaryEntry entry in copy)
            {
                bool shouldDrop = true;
                // scan through categories if_tdf threshold
                foreach (Category cat in categories)
                {
                    if (cat.TF_IDF.ContainsKey(entry.Key))
                        if ((double)cat.TF_IDF[entry.Key] >= cat.TF_IDF_Threshold)
                            shouldDrop = false;
                }
                if (shouldDrop)
                {
                    terms.Remove(entry.Key);
                    termsIDF.Remove(entry.Key);
                    DF.Remove(entry.Key);
                    foreach (Category cat in categories)
                    {
                        if (cat.Vocabulary.ContainsKey(entry.Key))
                            cat.DropTerm(entry.Key);
                    }
                }
                current++;
                if (BigMovementIndicated != null)
                {
                    BigMovementIndicated(current, Terms.Count);
                }
            }

            // re calculate condprob
            foreach (Category cat in categories)
            {
                cat.CondProb.Clear();
                cat.CondProbCalc(this.terms.Count);
            }
        }

        private void vocabularyMerger(Category cat)
        {
            foreach (System.Collections.DictionaryEntry term in cat.Vocabulary)
            {
                if (this.terms.ContainsKey(term.Key))
                {
                    // term exists in this.terms, increase frequency
                    this.terms[term.Key] = (int) this.terms[term.Key] + (int) term.Value;
                    this.DF[term.Key] = (int)this.DF[term.Key] + (int) cat.DocumentFreq[term.Key];
                }
                else
                {
                    // new term
                    this.terms.Add(term.Key,term.Value);
                    this.DF.Add(term.Key,cat.DocumentFreq[term.Key]);
                }
            }
        }

        public string Name
        {
            get { return this.name; }
        }

        public string Folder
        {
            get { return this.folder; }
        }

        public int CountDocuments
        {
            get { return this.countDocuments; }
        }

        public int CountCategories
        {
            get { return this.countCategories; }
        }

        public System.Collections.ArrayList Categories
        {
            get { return this.categories; }
        }

        public System.Collections.SortedList Terms
        {
            get { return this.terms; }
        }

        public System.Collections.SortedList TestFolder(string folder)
        {
            System.Collections.SortedList result = new System.Collections.SortedList();
            //each subdirectory is corresponding to 1 category
            string[] directories = Directory.GetDirectories(folder);
            int nClass = directories.Length;
            int i = 0;
            foreach (string directory in directories)
            {
                // name of the subfolder is the name of category
                string rightClass = directory.Substring(directory.LastIndexOf("\\") + 1);
                string[] filenames = Directory.GetFiles(directory);
                int[] nTrueFalse = new int[2];      // nTrueFalse[0]:number of right classification
                                                    // nTrueFalse[1]:number of right classification
                foreach (string filename in filenames)
                {
                    if (DocumentClass(filename).Name == rightClass)
                        nTrueFalse[0]++;
                    else nTrueFalse[1]++;
                }
                i++;
                if (BigMovementIndicated != null)
                    BigMovementIndicated(i, nClass);
                result.Add(rightClass, nTrueFalse);
            }
            return result;
        }

        public Category DocumentClass(string file)
        {
            TextReader tr = new StreamReader(file);
            StringBuilder text = new StringBuilder();
            string line;
            while ((line = tr.ReadLine())!=null)
            {
                text.Append(line + " ");
            }
            //Extract vocabulary 
            Tokenizer tokenizer = new Tokenizer();
            System.Collections.SortedList documentTokens = tokenizer.Partition(text.ToString());
            //Apply to each category, estimate score for each category
            double score;
            double max = double.MinValue;
            Category argMax = null;
            foreach (Category cat in this.categories)
            {
                score = Math.Log(cat.Prior);
                foreach (System.Collections.DictionaryEntry entry in documentTokens)
                {
                    if (this.terms.ContainsKey(entry.Key))
                    {
                        if (cat.CondProb.ContainsKey(entry.Key))
                            score += Math.Log((double)cat.CondProb[entry.Key]);
                        else
                            score += Math.Log(1 / cat.Denom);
                    }
                    // else: the entry is not in the model dictionary, skip it.
                }
                if (score > max)
                {
                    max = score;
                    argMax = cat;
                }
            }
            tr.Close();
            return argMax;
        }

        private void progressIndicated(int current, int max)
        {
            if (this.ProgressIndicated != null)
            {
                ProgressIndicated(current, max);
            }
        }

        private void idfCalc()
        {
            // convert from document frequency to IDF
            foreach (System.Collections.DictionaryEntry termDF in this.DF)
            {
                this.termsIDF.Add(termDF.Key, Math.Log(this.countDocuments / (1 + (int)termDF.Value)));
            }
        }

        private string folder;
        private string name;
        private int countDocuments = 0;
        private int countCategories = 0;
        private System.Collections.ArrayList categories = new System.Collections.ArrayList();
        private System.Collections.SortedList terms = new System.Collections.SortedList();
        private System.Collections.SortedList termsIDF = new System.Collections.SortedList();
        //document frequency
        private System.Collections.SortedList DF = new System.Collections.SortedList(); 
        private int thresholdMin = 0;
        private int thresholdMax = int.MaxValue;
    }
}
