﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.IO;
using System.Diagnostics;
using System.Threading;
using System.ComponentModel;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

namespace Text_Classification
{
    /// <summary>
    /// Interaction logic for Window1.xaml
    /// </summary>
    public partial class Window1 : Window
    {
        private Model model;
        private BackgroundWorker backgroundWorker;
        private BackgroundWorker backgroundTester;
        public Window1()
        {
            InitializeComponent();
            backgroundWorker =
                (BackgroundWorker)this.FindResource("backgroundWorker");
            backgroundTester =
                (BackgroundWorker)this.FindResource("backgroundTester");
        }

        private void button1_Click(object sender, RoutedEventArgs e)
        {
            System.Windows.Forms.FolderBrowserDialog folderList =
                       new System.Windows.Forms.FolderBrowserDialog();
            {
                folderList.Description = "Choose folder for build model";
                folderList.RootFolder = Environment.SpecialFolder.MyComputer;
                if (folderList.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                {
                    if (backgroundWorker.IsBusy != true)
                    {
                        backgroundWorker.RunWorkerAsync(folderList.SelectedPath);
                        //dowork -= buildModel;
                    }
                }
            }
        }

        private void chooseApplyFolder_Click(object sender, RoutedEventArgs e)
        {
            System.Windows.Forms.FolderBrowserDialog folderList =
                       new System.Windows.Forms.FolderBrowserDialog();
            {
                folderList.Description = "Choose folder for test";
                folderList.RootFolder = Environment.SpecialFolder.MyComputer;
                if (folderList.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                {
                    if (backgroundTester.IsBusy != true)
                    {
                        this.backgroundTester.RunWorkerAsync(
                            folderList.SelectedPath);
                            //@"C:\Classes\Spring 2011\CP SC 881 Machine Learning\Project\20news-bydate-test");
                    }
                }
            }
        }

        private void showProgress(int current, int max)
        {
            //progressBar1.Minimum = 0;
            
            backgroundWorker.ReportProgress(current * 100 / max);
        }

        private void showProgress2(int current, int max)
        {
            backgroundTester.ReportProgress(current * 100 / max);
        }

        private void updateProgressBar1(int current, int max)
        {
            progressBar1.Maximum = max;
            progressBar1.Value = current;
        }

        private void updateProgressBar2(int current, int max)
        {
            progressBar2.Maximum = max;
            progressBar2.Value = current;

        }

        private void BackgroundWorker_DoWork(object sender, System.ComponentModel.DoWorkEventArgs e)
        {
            model = new Model(e.Argument.ToString());
            model.BigMovementIndicated += showProgress;
            model.Build();
            model.BigMovementIndicated -= showProgress;
               
           // model.SetThreshold(10, 100); 
            // model.FinizeTerms(300);
        }

        private void BackgroundTester_DoWork(object sender, DoWorkEventArgs e)
        {
            StringBuilder text = new StringBuilder();
            model.BigMovementIndicated += showProgress2;
            for (int i = 1000; i > 0; i -= 450)
            {
                int nTrue = 0;
                int nFalse = 0;
                model.FinizeTerms(i);
                System.Collections.SortedList summary = model.TestFolder(e.Argument.ToString());
                text.AppendLine(string.Format("Test with {0} terms", model.Terms.Count));
                foreach (System.Collections.DictionaryEntry entry in summary)
                {
                    System.Int32[] eResult = (System.Int32[]) entry.Value;
                    text.AppendLine(string.Format("category {0,25} : {1,5} True | {2,5} False | {3,5} Total | {4,6:0.00} % correct"
                        ,entry.Key.ToString(), eResult[0], (int)eResult[1], (int)eResult[0]+ (int)eResult[1],
                        (double)eResult[0] * 100/((int)eResult[0] +(int)eResult[1])));
                    nTrue += (int)eResult[0];
                    nFalse += (int)eResult[1];
                }
                text.AppendLine(string.Format("In total, {0,6} True | {1,6} False | {2,6} in Total | {3,6:0.00} % correct",
                    nTrue, nFalse, nTrue + nFalse, (double)nTrue / (nTrue + nFalse) * 100));
                text.AppendLine();
            }
            model.BigMovementIndicated -= showProgress2;
        }

        // Open a file and serialize the model into binary format.
        private void saveModel(Model model, string filename)
        {
            Stream stream = File.Open(filename, FileMode.Create);
            BinaryFormatter bformatter = new BinaryFormatter();
            bformatter.Serialize(stream, model);
            stream.Close();

        }

        //Open the filename and read model from it.
        private Model openModel(string filename)
        {
            Model model;
            
            Stream stream = File.Open(filename, FileMode.Open);
            BinaryFormatter bformatter = new BinaryFormatter();
            model = (Model)bformatter.Deserialize(stream);
            stream.Close();
            return model;
        }

        private void BackgroundWorker_RunWorkerCompleted(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e)
        {
            textBox1.Text = "Number of categories:" + model.CountCategories.ToString();
            textBox1.Text += Environment.NewLine + "Total documents: " + model.CountDocuments;
            textBox1.Text += Environment.NewLine + "Categories list:";
            foreach (Category category in model.Categories)
            {
                textBox1.Text += Environment.NewLine + category.Name + " has " + category.CountDocuments;
            }

            textBox1.Text += Environment.NewLine;
            textBox1.Text += Environment.NewLine + "token list:";
            textBox1.Text += model.Terms.Count;
            //foreach (System.Collections.DictionaryEntry token in Category.Vocabulary())
            //{
            //    textBox1.Text += Environment.NewLine + "Token name:" + token.Key.ToString() + " has " +
            //                token.Value.ToString();
            //}
            applyFile.IsEnabled = true;
            save.IsEnabled = true;
            chooseApplyFolder.IsEnabled = true;
        }

        private void BackgroundTester_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
        }

        private void cancel_Click(object sender, RoutedEventArgs e)
        {
            if (backgroundWorker.WorkerSupportsCancellation)
            {
                backgroundWorker.CancelAsync();
            }
        }

        private void BackgroundWorker_ProgressChanged(object sender, System.ComponentModel.ProgressChangedEventArgs e)
        {
            progressBar1.Maximum = 100;
            progressBar1.Value = e.ProgressPercentage;
        }

        private void BackgroundTester_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            progressBar2.Maximum = 100;
            progressBar2.Value = e.ProgressPercentage;
        }

        private void applyFile_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                using (System.Windows.Forms.OpenFileDialog fileDialog =
                    new System.Windows.Forms.OpenFileDialog())
                {
                    fileDialog.CheckFileExists = true;
                    fileDialog.CheckPathExists = true;
                    fileDialog.RestoreDirectory = true;
                    fileDialog.Multiselect = true;
                    fileDialog.Title = "Open file to classify";
                    fileDialog.ValidateNames = true;
                    if (fileDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                    {
                        foreach (string filename in fileDialog.FileNames)
                        {
                            Category fileclass = model.DocumentClass(filename);
                            textBox1.Text += Environment.NewLine + fileclass.Name + " : " + filename;
                        }
                    }
                }                
            }
            catch (Exception ex)
            {
                MessageBox.Show("Cannot operate because: " + ex.Message);
            }
        }

        private void open_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                using (System.Windows.Forms.OpenFileDialog openModelDialog =
                new System.Windows.Forms.OpenFileDialog())
                {
                    openModelDialog.DefaultExt = "mdl";
                    openModelDialog.CheckFileExists = true;
                    openModelDialog.AddExtension = true;
                    openModelDialog.CheckFileExists = true;
                    openModelDialog.CheckPathExists = true;
                    openModelDialog.Title = "Open existing model";
                    openModelDialog.Multiselect = true;
                    if (openModelDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                    {
                        model = openModel(openModelDialog.FileName);
                        save.IsEnabled = true;
                        applyFile.IsEnabled = true;
                        chooseApplyFolder.IsEnabled = true;
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Cannot open model because: " + ex.Message);
            }
        }

        private void save_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                using (System.Windows.Forms.SaveFileDialog saveModelDialog =
                    new System.Windows.Forms.SaveFileDialog())
                {
                    saveModelDialog.CheckPathExists = true;
                    saveModelDialog.AddExtension = true;
                    saveModelDialog.ValidateNames = true;
                    saveModelDialog.CreatePrompt = true;
                    saveModelDialog.DefaultExt = "mdl";
                    saveModelDialog.Title = "Save model to file ...";
                    if (saveModelDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                    {
                        saveModel(model, saveModelDialog.FileName);
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Cannot save model because: " + ex.Message);
            }
        }        
    }
}
